<?php

/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */
// +----------------------------------------------------------------------+
// | Copyright (c) 2006 Jani Mikkonen                                     |
// +----------------------------------------------------------------------+
// | This source file is subject to the New BSD license, That is bundled  |
// | with this package in the file LICENSE, and is available through      |
// | the world-wide-web at                                                |
// | http://www.opensource.org/licenses/bsd-license.php                   |
// | If you did not receive a copy of the new BSDlicense and are unable   |
// | to obtain it through the world-wide-web, please send a note to       |
// | pajoye@php.net so we can mail you a copy immediately.                |
// +----------------------------------------------------------------------+
// | Author:  Jani Mikkonen <jani@mikkonen.info>                          |
// +----------------------------------------------------------------------+
//
/**
 * Specific validation methods for data used in Finland
 * 
 * @category   Validate
 * @package    Validate_FI
 * @author     Jani Mikkonen <jani@mikkonen.info>
 * @copyright  Copyright (C) 2006 Jani Mikkonen
 * @license    http://www.opensource.org/licenses/bsd-license.php  New BSD License
 * @version    CVS: $Id$
 * @link       http://pear.php.net/package/Validate_FI
 */

// {{{ class Validate_FI
/**
 * Data validation class for Finland
 *
 * This class provides methods to validate:
 * - Postal code
 * - Telephone number
 * - Car registration number
 * - Motorbike registration number
 * - Personal Identity number (HETU)
 * - Unique Identification number (SATU)
 * - Business ID (Y-tunnus)
 * - Finnish Party Identification number (OVT-tunnus)
 * - Value Added Tax number (ALV-numero)
 * - Finnish bank account number (tilinumero)
 * - Finnish bank reference number (viitenumero)
 * - Credit Card number
 * 
 * @category   Validate
 * @package    Validate_FI
 * @author     Jani Mikkonen <jani@mikkonen.info>
 * @license    http://www.opensource.org/licenses/bsd-license.php  New BSD License
 * @link       http://pear.php.net/package/Validate_FI
 */
class Validate_FI
{
    
    // {{{ bool Validate_FI::postalCode(string $code [, bool $strong = false])
    /**
     * Validate Finnish postal code
	 *
     * Five digit postal code, maybe with a leading 'FI-'
     *
     * @access      public
     * @param       string  $code the postal code to be validated
     * @param       bool    optional; strong checks 
     *                      (e.g. against a list of postal codes)
     * @return      bool    true if postal code is valid, false otherwise
     */
    function postalCode($code, $strong=false)
    {
        $preg  = "/^(FI-)?[0-9]{5}$/";
        $match = preg_match($preg, $code) ? true : false;
        return $match;
    }
    // }}}
    
    // {{{ bool Validate_FI::phoneNumber(string $number)
    /**
     * Validate Finnish telephone number
     *
     * Simple check: number must be numeric when (, ), -, +, ., ' ' chars removed
     *
     * @static
     * @access      public
     * @param       string  $number the telephone number to be validated
     * @return      bool    true if telephone number is valid, false otherwise
     */
    function phoneNumber($number){
        $number = str_replace(Array('(', ')', '-', '+', '.', ' '), '', $number);
        $preg   = "/^[0-9]{3,16}$/";
        $match  = (preg_match($preg, $number)) ? true : false;
        return $match;
    }
    // }}}
    
    // {{{ bool Validate_FI::carReg(string $reg)
    /**
     * Validate Finnish car registration number
     *
     * Format: AAA-XXX
     *
     * Where AAA is 2-3 letters UPPERCASE A-Z and XXX is 1-3 numbers.
     *
     * @access      public
     * @param       string  $reg the registration number to be validated
     * @return      bool    true if registration number is valid, false otherwise
     */
    function carReg($reg){
        $preg  = "/^[A-Z]{2,3}-[0-9]{1,3}$/";
        $match = (preg_match($preg, $reg)) ? true : false;
        return $match;
    }
    // }}}
    
    // {{{ bool Validate_FI::bikeReg(string $reg)
    /**
     * Validate Finnish  motorbike registration number
     *
     * Format: AAAXXX
     *
     * Where AAA is 2-3 letters UPPERCASE A-Z and XXX is 1-3 numbers.
     * Letters and numbers are actually in separate lines.
     * 
     * @access      public
     * @param       string  $reg the registration number to be validated
     * @return      bool    true if registration number is valid, false otherwise
     */
    function bikeReg($reg){
        $preg = "/^[A-Z]{2,3}(\n)?[0-9]{1,3}$/";
        $match = (preg_match($preg, $reg)) ? true : false;
        return $match;
    }
    // }}}
    
    // {{{ mixed Validate_FI::pin(string $pin [, bool $gender = false])
    /**
     * Validate Personal Identity Number (HETU)
     *
     * The Finnish PIN (HETU) number is a 11 digit number with birthdate as
     * ddmmyycxxxy where c is century, xxx is a three digit individual number
     * and the last digit is a control number (y). 
     * 
     * If xxx is odd it's a male PIN number and if even a female.
     * 
     * Return gender (M or F) if valid PIN, is available by
     * switching $gender (2nd parameter) to true.
     * 
     * Example: 010101-123N would be a male and born in 1st of January 1901.
     * 
     * More information in Finnish:
     *
     * @link http://koti.mbnet.fi/~thales/tarkmerk.htm#hetu1
     *
     * @access      public
     * @param       string  $pin PIN number to be validated
     * @param       bool    optional; Returns gender (M or F) if valid PIN
     * @return      mixed   Returns true or false without gender check,
     *                      and M or F if $gender = true and PIN is valid.
     */
    function pin($pin, $gender = false)
    {
        $pin            = strtoupper($pin);
        static $control = array("0","1","2","3","4","5","6","7","8","9",
                                "A","B","C","D","E","F","H","J","K","L",
                                "M","N","P","R","S","T","U","V","W","X","Y");
        static $century = array('+' => "18",
                                '-' => "19",
                                'A' => "20");
        if(ereg("^([0-9]{2})([0-9]{2})([0-9]{2})([+-A]{1})([0-9]{3})([0-9A-Z]{1})$",$pin,$regs)) {
            // Validate a Gregorian date
            if(checkdate($regs[2],$regs[1],$century[$regs[4]].$regs[3])) {
                if($control[intval($regs[1].$regs[2].$regs[3].$regs[5])%31] == $regs[6]) {
                    if($gender) {
                        return (($regs[5]%2==1)) ? 'M' : 'F';
                    } else {
                        return true;
                    }
                }
            } else {
                return false;
            }
        } else { 
            return false;
        }
    }
    // }}}
    
    // {{{ bool Validate_FI::finuid(string $finuid)
    /**
     * Validate Finnish Unique Identification Number (SATU)
     *
     * The Finnish FINUID (SATU) number is a 9 digit number.
     * The last digit is a control number (y).
     * Example: 10011187H
     * 
     * More information in Finnish:
     *
     * @link http://koti.mbnet.fi/~thales/tarkmerk.htm#satu
     *
     * @access      public
     * @param       string  $finuid FINUID number to be validated
     * @return      bool    true if FINUID is valid, false otherwise
     */
    function finuid($finuid)
    {
        $finuid         = strtoupper($finuid);
        static $control = array("0","1","2","3","4","5","6","7","8","9",
                                "A","B","C","D","E","F","H","J","K","L",
                                "M","N","P","R","S","T","U","V","W","X","Y");
        if(ereg("^([0-9]{8})([0-9A-Z]{1})$",$finuid,$regs)) {
            if($control[intval($regs[1])%31] == $regs[2]) {
                return true;
            }
        } else { 
            return false;
        }
    }
    // }}}
    
    // {{{ bool Validate_FI::businessId(string $businessid)
    /**
     * Validate Finnish Business ID (Y-tunnus)
     *
     * The Finnish Business ID (Y-tunnus) number is a 9 digit number.
     * The last digit is a control number (y).
     * Example: xxxxxxx-y
     * 
     * More information in Finnish:
     *
     * @link http://koti.mbnet.fi/~thales/tarkmerk.htm#y-tunnus2
     *
     * @access      public
     * @param       string  $businessid Business ID number to be validated
     * @return      bool    true if Business ID is valid, false otherwise
     */
    function businessId($businessid)
    {
        $prepreg = "/^[0-9]{6,7}-[0-9]{1}$/";
        if (preg_match($prepreg, $businessid)) {
            list($number, $control) = split('[-]', $businessid);
            // Add leading zeros if number is < 7
            $number = str_pad($number, 7, 0, STR_PAD_LEFT);
            $controlSum = 0;
            $controlSum += (int)substr($number, 0, 1)*7;
            $controlSum += (int)substr($number, 1, 1)*9;
            $controlSum += (int)substr($number, 2, 1)*10;
            $controlSum += (int)substr($number, 3, 1)*5;
            $controlSum += (int)substr($number, 4, 1)*8;
            $controlSum += (int)substr($number, 5, 1)*4;
            $controlSum += (int)substr($number, 6, 1)*2;
            $controlSum = $controlSum%11;
            if($controlSum == 0) {
                return ($controlSum == $control) ? true : false;
            } else if($controlSum >= 2 && $controlSum <= 10 ) {
                return ((11 - $controlSum) == $control) ? true : false;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }
    // }}}
    
    // {{{ bool Validate_FI::partyId(int $partyid)
    /**
     * Validate Finnish Party Identification (OVT-tunnus)
     *
     * The Finnish Party Identification (OVT-tunnus) number is a 12-17 digit number
     * and it is generated from Business ID.
     * 
     * Example: 0037AAAAAAAABBBBB, 0037 indicates Finland, AAAAAAAA is the 
     * Business ID and BBBBB is optional organization number.
     * 
     * More information in Finnish:
     *
     * @link http://koti.mbnet.fi/~thales/tarkmerk.htm#alv-numero
     *
     * @access      public
     * @param       int     $partyid Party Identification number to be validated
     * @return      bool    true if partyid is valid, false otherwise
     * @see         businessId()
     */
    function partyId($partyid)
    {
        $prepreg = "/^[0-9]{12,17}$/";
        if (preg_match($prepreg, $partyid)) {
            $countryCode = substr($partyid, 0, 4);
            $controlNum  = substr($partyid, 11, 1);
            $businessNum = substr($partyid, 4, 7);
            $businessId  = $businessNum.'-'.$controlNum;
            if($countryCode == '0037' && Validate_FI::businessId($businessId)) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }
    // }}}
    
    // {{{ bool Validate_FI::vatNumber(string $vatnumber)
    /**
     * Validate Finnish Value Added Tax number (ALV-numero)
     *
     * The Finnish VAT number (ALV-numero) is maximum 
     * of 14 digits and is generated from Business ID.
     * 
     * More information in Finnish:
     *
     * @link http://koti.mbnet.fi/~thales/tarkmerk.htm#alv-numero
     *
     * @access      public
     * @param       string  $vatnumber VAT number to be validated
     * @return      bool    true if VAT number is valid, false otherwise
     * @see         businessId()
     */
    function vatNumber($vatnumber)
    {
        $countryCode = substr($vatnumber, 0, 2);
        $controlNum  = substr($vatnumber, -1, 1);
        $businessNum = substr($vatnumber, 2, -1);
        $businessId  = $businessNum.'-'.$controlNum;
        if($countryCode == 'FI' && Validate_FI::businessId($businessId)) {
            return true;
        } else {
            return false;
        }
    }
    // }}}
    
    // {{{ bool Validate_FI::bankAccount(string $number)
    /**
     * Validate Finnish bank account number
     * 
     * This method checks the bank account number according to 
     * Finnish Bank Association.
     * 
     * Format: XXXXXX-XXXXXXXX, 6 digits, - and 2-8 digits.
     * 
     * More information in Finnish:
     *
     * @link http://koti.mbnet.fi/~thales/tarkmerk.htm#pankkitili
     * 
     * @access      public
     * @param       string  $number Finnish bank account number to be validated
     * @return      bool    true if bank account is valid, false otherwise
     * @see         mod10()
     */
    function bankAccount($number)
    {
        $prepreg = "/^[0-9]{6}-[0-9]{2,8}$/";
        if (preg_match($prepreg, $number)) {
            // Bank groups are identified by the first digit
            $bankType = substr($number, 0, 1);
            // Group 1: First digit is 1, 2, 3, 6 or 8
            $bankGroup1 = array('1','2','3','6','8');
            // Group 2: First digit is 4 or 5
            $bankGroup2 = array('4','5');
            // split account number
            ereg("([0-9]{6})-([0-9]{2,8})", $number, $regs);
            if (in_array($bankType, $bankGroup1)) {
                // Group 1: 999999-99999 -> 999999-00099999
                $number = $regs[1] . str_pad($regs[2], 8, 0, STR_PAD_LEFT);
                // Group 2: 999999-99999 -> 999999-90009999
            } else if (in_array($bankType, $bankGroup2)) {
                $number = $regs[1] . substr($regs[2], 0, 1) . 
                          str_pad(substr($regs[2], 1, 7), 7, 0, STR_PAD_LEFT);
            }
            // Now when we have a 14 digit bank account number, we can validate it.
            return Validate_FI::_mod10($number);
        } else {
            return false;
        }
    }
    // }}}
    
    // {{{ bool Validate_FI::refNum(string $refnum)
    /**
     * Validate Finnish bank reference number
     * 
     * This method checks the bank reference number according to 
     * Finnish Bank Association.
     * 
     * More information in Finnish:
     *
     * @link http://koti.mbnet.fi/~thales/tarkmerk.htm#viitenumero
     * 
     * @access      public
     * @param       string  $refnum Finnish bank reference number to be validated
                            (spaces and dashes tolerated)
     * @return      bool    true if reference number is valid, false otherwise
     * @see         mod10()
     */
    function refNum($refnum)
    {
        // Remove non-numeric characters from $refnum. Only 4-20 digit number.
        $refnum = ereg_replace('[^0-9]+', '', $refnum);
        // The last digit is a control number.
        $controlNum  = substr($refnum, -1, 1);
        // Subtract control number from the $refnum.
        $refNumCheck = substr($refnum, 0, strlen($refnum)-1);
        $prepreg = "/^[0-9]{3,19}$/";
        if(preg_match($prepreg, $refNumCheck)) {
            // remove leading zeros
            $refNumCheck = ltrim($refNumCheck, 0);
            $mul = 7;
            $refSum = 0;
            for($refLength = strlen($refNumCheck); $refLength > 0; $refLength--) {
                $refSum += substr($refNumCheck, $refLength - 1, 1) * $mul;
                switch ($mul) {
                case 7:
                    $mul = 3;
                    break;
                case 3:
                    $mul = 1;
                    break;
                case 1:
                    $mul = 7;
                    break;
                }
            }
            $refSum = substr(10 - ($refSum % 10), -1);
            if($refSum == $controlNum){
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }
    // }}}
    
    // {{{ bool Validate_FI::creditCard(string $number)
    /**
     * Validate credit card number
     *
     * This method checks the credit number according to Luhn algorithm.
     * This method doesn't guarantee that the card is legitimate.
     *
     * @access      public
     * @param       string  $number credit card number to be validated
     *                      (spaces and dashes tolerated)
     * @return      bool    true if credit card number is valid, false otherwise
     * @see         mod10()
     */
    function creditCard($number)
    {
        // Remove non-numeric characters from $number 
        $number = ereg_replace('[^0-9]+', '', $number);
        // Validate number
        return Validate_FI::_mod10($number);
    }
    // }}}
    
    // {{{ bool Validate_FI::_mod10(string $number)
    /**
     * Validate number according to Luhn algorithm (mod 10)
     *
     * This method checks given number according Luhn algorithm.
     * See Wikipedia for more information:
     *
     * @link http://en.wikipedia.org/wiki/Luhn_algorithm
     *
     * @access      private
     * @param       string  $number to be validated
     * @return      bool    true if number is valid, false otherwise
     */
    function _mod10($number)
    {
        // Double every second digit started at the right
        $doubledNumber  = "";
        $odd            = false;
        for($i = strlen($number)-1; $i >=0; $i--)
        {
            $doubledNumber .= ($odd) ? $number[$i]*2 : $number[$i];
            $odd            = !$odd;
        }
        // Add up each 'single' digit
        $sum = 0;
        for($i = 0; $i < strlen($doubledNumber); $i++) {
            $sum += (int)$doubledNumber[$i];
        }
        // A valid number doesn't have a remainder after mod10 or equal to 0
        return (($sum % 10 == 0) && ($sum != 0)) ? true : false;
    }
    // }}}
}
// }}}

?>
